__author__ = 'jatwood'

import lasagne
import lasagne.layers as layers
import theano
import theano.tensor as T
import numpy as np
import matplotlib.pyplot as plt

import util


# This class is not user facing; it contains the Lasagne internals for the SCNN model.
class SearchConvolution(layers.MergeLayer):
    """
    A search-convolutional Lasagne layer.
    """
    def __init__(self, incomings, n_hops, n_features,
                 W=lasagne.init.Normal(0.01),
                 nonlinearity=lasagne.nonlinearities.tanh,
                 **kwargs):
        super(SearchConvolution, self).__init__(incomings, **kwargs)

        self.W = self.add_param(W, (n_hops,n_features), name='W')

        self.n_hops = n_hops
        self.n_features = n_features

        self.nonlinearity = nonlinearity

    def get_output_for(self, inputs, **kwargs):
        """
        Compute search convolution of inputs.

        :param inputs: [Apow, X]
        :return: Search convolution of inputs with shape (self.nhops, self.nfeatures)
        """

        Apow = inputs[0]
        X = inputs[1]

        def compute_output(i, w, a, x, h):
            """

            :param i: index
            :param w: weight vector (n_features,)
            :param x: feature vector (n_nodes, n_features)
            :param h: n_hops
            :param a: adjacency matrix (n_nodes, n_nodes)
            :return: output[i]
            """
            return (T.dot(a, x).transpose()) * T.addbroadcast(T.reshape(w, (w.shape[0],1)),1)

        seq_values = np.arange(self.n_hops)
        seq = theano.shared(value = seq_values, name="seq", borrow=True)

        out, _ = theano.scan(fn=compute_output,
                             non_sequences=[X, self.n_hops],
                             sequences=[seq, self.W, Apow],
                             n_steps = self.n_hops)

        return self.nonlinearity(out.transpose())

    def get_output_shape_for(self, input_shapes):
        return (input_shapes[0][0], self.n_hops, self.n_features)


# This class is user-facing.  It contains a full SCNN model.
class SCNN:
    """
    The search-convolutional neural network model.
    """
    def __init__(self, n_hops=2, transform_fn=util.rw_laplacian):
        self.n_hops = n_hops
        self.transform_fn = transform_fn

        # Initialize Theano variables
        self.var_A = T.matrix('A')
        self.var_Apow = T.tensor3('Apow')
        self.var_X = T.matrix('X')
        self.var_Y = T.imatrix('Y')

    def fit(self, A, X, Y, train_indices, valid_indices,
            learning_rate=0.05, batch_size=100, n_epochs=100,
            loss_fn=lasagne.objectives.multiclass_hinge_loss,
            update_fn=lasagne.updates.adagrad,
            stop_early=True,
            stop_window_size=5,
            output_weights=False,
            show_weights=False):

        # Ensure that data have the correct dimensions
        assert A.shape[0] == X.shape[0]
        assert X.shape[0] == Y.shape[0]
        assert len(Y.shape) > 1

        if self.transform_fn is not None:
            A = self.transform_fn(A)

        # Extract dimensions
        n_nodes = A.shape[0]
        n_features = X.shape[1] + 1
        n_classes = Y.shape[1]

        n_batch = n_nodes // batch_size

        # Compute the matrix power series
        Apow = util.A_power_series(A, self.n_hops)

        # Add bias term to X
        X = np.hstack([X, np.ones((X.shape[0],1))]).astype('float32')

        # Create Lasagne layers
        self.l_in_apow = lasagne.layers.InputLayer((self.n_hops, batch_size, n_nodes), input_var=self.var_Apow)
        self.l_in_x = lasagne.layers.InputLayer((n_nodes, n_features), input_var=self.var_X)
        self.l_sc = SearchConvolution([self.l_in_apow, self.l_in_x], self.n_hops + 1, n_features)
        self.l_out = layers.DenseLayer(self.l_sc, num_units=n_classes, nonlinearity=lasagne.nonlinearities.tanh)

        # Create symbolic representations of predictions, loss, parameters, and updates.
        prediction = layers.get_output(self.l_out)
        loss = lasagne.objectives.aggregate(loss_fn(prediction, self.var_Y), mode='mean')
        params = lasagne.layers.get_all_params(self.l_out)
        updates = update_fn(loss, params, learning_rate=learning_rate)

        # Create functions that apply the model to data and return loss
        apply_loss = theano.function([self.var_Apow, self.var_X, self.var_Y],
                                      loss, updates=updates)

        # Train the model
        print 'Training model...'
        validation_losses = []
        validation_loss_window = np.zeros(stop_window_size)
        validation_loss_window[:] = float('+inf')

        for epoch in range(n_epochs):
            train_loss = 0.0

            np.random.shuffle(train_indices)

            for batch in range(n_batch):
                start = batch * batch_size
                end = min((batch + 1) * batch_size, train_indices.shape[0])

                if start < end:
                    train_loss += apply_loss(Apow[:,train_indices[start:end],:],
                                             X,
                                             Y[train_indices[start:end],:])

            valid_loss = apply_loss(Apow[:,valid_indices,:],
                                    X,
                                    Y[valid_indices,:])

            print "Epoch %d training error: %.6f" % (epoch, train_loss)
            print "Epoch %d validation error: %.6f" % (epoch, valid_loss)

            validation_losses.append(valid_loss)

            if output_weights:
                W = layers.get_all_param_values(self.l_sc)[0]
                np.savetxt('W_%d.csv' % (epoch,), W, delimiter=',')

            if show_weights:
                W = layers.get_all_param_values(self.l_sc)[0]
                plt.imshow(W, aspect='auto', interpolation='none')
                plt.show()

            if stop_early:
                if valid_loss >= validation_loss_window.mean():
                    print 'Validation loss did not decrease. Stopping early.'
                    break
            validation_loss_window[epoch % stop_window_size] = valid_loss

    def predict(self, A, X, test_indices):
        if self.transform_fn is not None:
            A = self.transform_fn(A)

        # Compute the matrix power series
        Apow = util.A_power_series(A, self.n_hops)

        # add bias term to X
        X = np.hstack([X, np.ones((X.shape[0],1))]).astype('float32')

        # Create symbolic representation of predictions
        pred = layers.get_output(self.l_out)

        # Create a function that applies the model to data to predict a class
        pred_fn = theano.function([self.var_Apow, self.var_X], T.argmax(pred, axis=1), allow_input_downcast=True)

        # Return the predictions
        predictions = pred_fn(Apow[:,test_indices,:], X)
        return predictions

    def predict_proba(self, A, X, test_indices):
        if self.transform_fn is not None:
            A = self.transform_fn(A)

        # Compute the matrix power series
        Apow = util.A_power_series(A, self.n_hops)

        # add bias term to X
        X = np.hstack([X, np.ones((X.shape[0],1))]).astype('float32')

        # Create symbolic representation of predictions
        pred = layers.get_output(self.l_out)

        # Create a function that applies the model to data to predict a class
        pred_fn = theano.function([self.var_Apow, self.var_X], (pred - pred.min(axis=1,keepdims=True)) / (pred - pred.min(axis=1,keepdims=True)).sum(axis=1,keepdims=True), allow_input_downcast=True)

        # Return the predictions
        predictions = pred_fn(Apow[:,test_indices,:], X)
        return predictions



__author__ = 'jatwood'

import numpy as np

import data
from scnn import SCNN

def test_scnn():
    from sklearn.metrics import f1_score

    A, X, Y = data.parse_cora()

    n_nodes = A.shape[0]

    indices = np.arange(n_nodes)
    train_indices = indices[:n_nodes // 3]
    valid_indices = indices[n_nodes // 3:(2* n_nodes) // 3]
    test_indices  = indices[(2* n_nodes) // 3:]

    scnn = SCNN(n_hops=2)
    scnn.fit(A, X, Y, train_indices=train_indices, valid_indices=valid_indices)

    probs = scnn.predict_proba(A, X, test_indices)
    print probs

    preds = scnn.predict(A, X, test_indices)
    actuals = np.argmax(Y[test_indices,:], axis=1)

    print 'F score: %.4f' % (f1_score(actuals, preds))

if __name__ == '__main__':
    import nose
    nose.run(defaultTest=__name__, argv=['.','-s'])



__author__ = 'jatwood'

import numpy as np
import cPickle as cp


def parse_cora(plot=False):
    path = "data/cora/"

    id2index = {}

    label2index = {
        'Case_Based': 0,
        'Genetic_Algorithms': 1,
        'Neural_Networks': 2,
        'Probabilistic_Methods': 3,
        'Reinforcement_Learning': 4,
        'Rule_Learning': 5,
        'Theory': 6
    }

    features = []
    labels = []

    with open(path + 'cora.content', 'r') as f:
        i = 0
        for line in f.xreadlines():
            items = line.strip().split('\t')

            id = items[0]

            # 1-hot encode labels
            label = np.zeros(len(label2index))
            label[label2index[items[-1]]] = 1
            labels.append(label)

            # parse features
            features.append([int(x) for x in items[1:-1]])

            id2index[id] = i
            i += 1

    features = np.asarray(features, dtype='float32')
    labels = np.asarray(labels, dtype='int32')

    n_papers = len(id2index)

    adj = np.zeros((n_papers, n_papers), dtype='float32')

    with open(path + 'cora.cites', 'r') as f:
        for line in f.xreadlines():
            items = line.strip().split('\t')
            adj[ id2index[items[0]], id2index[items[1]] ] = 1.0
            # undirected
            adj[ id2index[items[1]], id2index[items[0]] ] = 1.0

    if plot:
        import networkx as nx
        import matplotlib.pyplot as plt

        #G = nx.from_numpy_matrix(adj, nx.DiGraph())
        G = nx.from_numpy_matrix(adj, nx.Graph())
        print G.order()
        print G.size()
        plt.figure()
        nx.draw(G, node_size=10, edge_size=1)
        plt.savefig('../data/cora_net.pdf')

        plt.figure()
        plt.imshow(adj)
        plt.savefig('../data/cora_adj.pdf')

    return adj.astype('float32'), features.astype('float32'), labels.astype('int32')


def parse_nci(graph_name='nci1.graph'):
    path = "data/nci/"

    with open(path+graph_name,'r') as f:
        raw = cp.load(f)

        n_classes = 2
        n_graphs = len(raw['graph'])

        A = []
        rX = []
        Y = np.zeros((n_graphs, n_classes), dtype='int32')

        for i in range(n_graphs):
            # Set label
            Y[i][raw['labels'][i]] = 1

            # Parse graph
            G = raw['graph'][i]

            n_nodes = len(G)

            a = np.zeros((n_nodes,n_nodes), dtype='float32')
            x = np.zeros((n_nodes,1), dtype='int32')

            for node, meta in G.iteritems():
                x[node,0] = meta['label'][0]
                for neighbor in meta['neighbors']:
                    a[node, neighbor] = 1

            A.append(a)
            rX.append(x)

        maxval = max([max(x) for x in rX])

        X = [np.zeros((rx.size, maxval+1), dtype='float32') for rx in rX]
        for i in range(len(X)):
            X[i][np.arange(rX[i].size),rX[i]] = 1

    return A, X, Y




__author__ = 'jatwood'


# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

setup(
    name='scnn',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='1.2.0',

    description='Search-Convolutional Neural Networks',
    long_description="""
SCNN
====

An implementation of search-convolutional neural networks [1], a new model for graph-structured data.

Installation
------------
Using pip:

    pip install scnn

Usage
-----

	import numpy as np
    from scnn import SCNN, data
    from sklearn.metrics import f1_score

	# Parse the cora dataset and return an adjacency matrix, a design matrix, and a 1-hot label matrix
    A, X, Y = data.parse_cora()

	# Construct array indices for the training, validation, and test sets
    n_nodes = A.shape[0]
    indices = np.arange(n_nodes)
    train_indices = indices[:n_nodes // 3]
    valid_indices = indices[n_nodes // 3:(2* n_nodes) // 3]
    test_indices  = indices[(2* n_nodes) // 3:]

	# Instantiate an SCNN and fit it to cora
    scnn = SCNN()
    scnn.fit(A, X, Y, train_indices=train_indices, valid_indices=valid_indices)

	# Predict labels for the test set
    preds = scnn.predict(A, X, test_indices)
    actuals = np.argmax(Y[test_indices,:], axis=1)

	# Display performance
    print 'F score: %.4f' % (f1_score(actuals, preds))

References
----------

[1] http://arxiv.org/abs/1511.02136
    """,

    # The project's main homepage.
    url='https://github.com/jcatw/scnn',

    # Author details
    author='James Atwood',
    author_email='james.c.atwood@gmail.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],

    # What does your project relate to?
    keywords='search convolutional neural networks machine learning',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['numpy',
                      'theano',
                      'lasagne>=0.1dev',
                      'matplotlib',
                      ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={},

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={'scnn': ['README.md','scnn/data/cora', 'scnn/data/nci']},

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={},
)



__author__ = 'jatwood'

import numpy as np

def rw_laplacian(A):
    Dm1 = np.zeros(A.shape)

    degree = A.sum(0)

    for i in range(A.shape[0]):
        if degree[i] == 0:
            Dm1[i,i] = 0.
        else:
            Dm1[i,i] = - 1. / degree[i]

    return -np.asarray(Dm1.dot(A),dtype='float32')

def A_power_series(A,k):
    """
    Computes [A**0, A**1, ..., A**k]

    :param A: 2d numpy array
    :param k: integer, degree of series
    :return: 3d numpy array [A**0, A**1, ..., A**k]
    """
    assert k >= 0

    Apow = [np.identity(A.shape[0])]

    if k > 0:
        Apow.append(A)

        for i in range(2, k+1):
            Apow.append(np.dot(A, Apow[-1]))

    return np.asarray(Apow, dtype='float32')



__author__ = 'jatwood'

import numpy as np
import lasagne

import data
from graph_scnn import GraphSCNN

def test_graph_scnn():
    from sklearn.metrics import accuracy_score
    A, X, Y = data.parse_nci()

    n_graphs = len(A)

    indices = np.arange(n_graphs)
    np.random.shuffle(indices)

    train_indices = indices[:n_graphs // 3]
    valid_indices = indices[n_graphs // 3:(2* n_graphs) // 3]
    test_indices  = indices[(2* n_graphs) // 3:]

    scnn = GraphSCNN(n_hops=5)
    scnn.fit(A, X, Y, train_indices=train_indices, valid_indices=valid_indices,
             learning_rate=0.05, batch_size=100, n_epochs=500)


    preds = scnn.predict(A, X, test_indices)
    actuals = np.argmax(Y[test_indices,:], axis=1)

    print 'Accuracy: %.4f' % (accuracy_score(actuals, preds))

if __name__ == '__main__':
    import nose
    nose.run(defaultTest=__name__, argv=['.','-s'])
